import os
import datetime
from logging import getLogger
from copy import deepcopy

from abc import ABC
from collections.abc import MutableMapping

from jsonschema import validate
from yacman import YAMLConfigManager, select_config

from typing import Optional, Union, Dict, Any, List, Iterator


from .exceptions import (
    PipestatDependencyError,
    NoBackendSpecifiedError,
    SchemaNotFoundError,
    InvalidTimeFormatError,
    PipestatDatabaseError,
    RecordNotFoundError,
)
from pipestat.backends.file_backend.filebackend import FileBackend
from .reports import HTMLReportBuilder, _create_stats_objs_summaries
from .helpers import validate_type, mk_abs_via_cfg, read_yaml_data, default_formatter, zip_report
from .const import (
    PKG_NAME,
    DEFAULT_PIPELINE_NAME,
    ENV_VARS,
    CFG_DATABASE_KEY,
    SCHEMA_PATH,
    STATUS_SCHEMA,
    STATUS_SCHEMA_SOURCE_KEY,
    STATUS_SCHEMA_KEY,
    STATUS_FILE_DIR,
    FILE_KEY,
    DB_ONLY_KEY,
    DB_URL,
    PIPELINE_NAME,
    PIPELINE_TYPE,
    PROJECT_NAME,
    RECORD_IDENTIFIER,
    RESULT_FORMATTER,
    MULTI_PIPELINE,
    OUTPUT_DIR,
    CREATED_TIME,
    MODIFIED_TIME,
    CFG_SCHEMA,
    CONFIG_KEY,
    SCHEMA_KEY,
    SAMPLE_NAME_ID_KEY,
    DATA_KEY,
)

try:
    from pipestat.backends.db_backend.db_parsed_schema import (
        ParsedSchemaDB as ParsedSchema,
    )
except ImportError:
    from .parsed_schema import ParsedSchema

try:
    from pipestat.backends.db_backend.dbbackend import DBBackend
    from pipestat.backends.db_backend.db_helpers import construct_db_url
except ImportError:
    # We let this pass, but if the user attempts to create DBBackend, check_dependencies raises exception.
    pass


_LOGGER = getLogger(PKG_NAME)


def check_dependencies(dependency_list: list = None, msg: str = None):
    """Decorator to check that the dependency list has successfully been imported."""

    def wrapper(func):
        def inner(*args, **kwargs):
            dependencies_satisfied = True
            if dependency_list is not None:
                for i in dependency_list:
                    if i not in globals():
                        _LOGGER.warning(msg=f"Missing dependency: {i}")
                        dependencies_satisfied = False
            if dependencies_satisfied is False:
                raise PipestatDependencyError(msg=msg)
            return func(*args, **kwargs)

        return inner

    return wrapper


def require_backend(func):
    """Decorator to ensure a backend exists for functions that require one"""

    def inner(self, *args, **kwargs):
        if not self.backend:
            raise NoBackendSpecifiedError
        return func(self, *args, **kwargs)

    return inner


class PipestatManager(MutableMapping):
    """
    Pipestat standardizes reporting of pipeline results and
    pipeline status management. It formalizes a way for pipeline developers
    and downstream tools developers to communicate -- results produced by a
    pipeline can easily and reliably become an input for downstream analyses.
    A PipestatManager object exposes an API for interacting with the results and
    pipeline status and can be backed by either a YAML-formatted file
    or a database.
    """

    def __init__(
        self,
        project_name: Optional[str] = None,
        record_identifier: Optional[str] = None,
        schema_path: Optional[str] = None,
        results_file_path: Optional[str] = None,
        database_only: Optional[bool] = True,
        config_file: Optional[str] = None,
        config_dict: Optional[dict] = None,
        flag_file_dir: Optional[str] = None,
        show_db_logs: bool = False,
        pipeline_type: Optional[str] = None,
        pipeline_name: Optional[str] = DEFAULT_PIPELINE_NAME,
        result_formatter: staticmethod = default_formatter,
        multi_pipelines: bool = False,
        output_dir: Optional[str] = None,
    ):
        """
        Initialize the PipestatManager object

        :param str record_identifier: record identifier to report for. This
            creates a weak bound to the record, which can be overridden in
            this object method calls
        :param str schema_path: path to the output schema that formalizes
            the results structure
        :param str results_file_path: YAML file to report into, if file is
            used as the object back-end
        :param bool database_only: whether the reported data should not be
            stored in the memory, but only in the database
        :param str config_file: path to the configuration file
        :param dict config_dict:  a mapping with the config file content
        :param str flag_file_dir: path to directory containing flag files
        :param bool show_db_logs: Defaults to False, toggles showing database logs
        :param str pipeline_type: "sample" or "project"
        :param str pipeline_name: name of the current pipeline, defaults to
        :param str result_formatter: function for formatting result
        :param bool multi_pipelines: allows for running multiple pipelines for one file backend
        :param str output_dir: target directory for report generation via summarize and table generation via table.
        """

        super(PipestatManager, self).__init__()

        # Initialize the cfg dict as an attribute that holds all configuration data
        self.cfg = {}

        # Load and validate database configuration
        # If results_file_path exists, backend is a file else backend is database.

        self.cfg["config_path"] = select_config(config_file, ENV_VARS["config"])

        self.cfg[CONFIG_KEY] = YAMLConfigManager(
            entries=config_dict, filepath=self.cfg["config_path"]
        )

        _, cfg_schema = read_yaml_data(CFG_SCHEMA, "config schema")
        validate(self.cfg[CONFIG_KEY].exp, cfg_schema)

        self.cfg[SCHEMA_PATH] = self.cfg[CONFIG_KEY].priority_get(
            "schema_path", env_var=ENV_VARS["schema"], override=schema_path
        )
        self.process_schema(schema_path)

        self.cfg[RECORD_IDENTIFIER] = self.cfg[CONFIG_KEY].priority_get(
            "record_identifier", env_var=ENV_VARS["record_identifier"], override=record_identifier
        )

        self.cfg[PIPELINE_NAME] = (
            self.cfg[SCHEMA_KEY].pipeline_name
            if self.cfg[SCHEMA_KEY] is not None
            else pipeline_name
        )

        self.cfg[PROJECT_NAME] = self.cfg[CONFIG_KEY].priority_get(
            "project_name", env_var=ENV_VARS["project_name"], override=project_name
        )

        self.cfg[SAMPLE_NAME_ID_KEY] = self.cfg[CONFIG_KEY].priority_get(
            "record_identifier",
            env_var=ENV_VARS["sample_name"],
            override=record_identifier,
        )

        self.cfg[DB_ONLY_KEY] = database_only

        self.cfg[PIPELINE_TYPE] = self.cfg[CONFIG_KEY].priority_get(
            "pipeline_type", default="sample", override=pipeline_type
        )

        self.cfg[FILE_KEY] = mk_abs_via_cfg(
            self.resolve_results_file_path(
                self.cfg[CONFIG_KEY].priority_get(
                    "results_file_path",
                    env_var=ENV_VARS["results_file"],
                    override=results_file_path,
                )
            ),
            self.cfg["config_path"],
        )

        self.cfg[RESULT_FORMATTER] = result_formatter

        self.cfg[MULTI_PIPELINE] = multi_pipelines

        self.cfg["multi_result_files"] = None

        self.cfg[OUTPUT_DIR] = self.cfg[CONFIG_KEY].priority_get("output_dir", override=output_dir)

        if self.cfg[FILE_KEY]:
            self.initialize_filebackend(record_identifier, results_file_path, flag_file_dir)

        else:
            self.initialize_dbbackend(record_identifier, show_db_logs)

    def __str__(self):
        """
        Generate string representation of the object

        :return str: string representation of the object
        """
        res = f"{self.__class__.__name__} ({self.cfg[PIPELINE_NAME]})"
        res += "\nBackend: {}".format(
            f"File\n - results: {self.cfg[FILE_KEY]}\n - status: {self.cfg[STATUS_FILE_DIR]}"
            if self.cfg[FILE_KEY]
            else f"Database (dialect: {self.backend.db_engine_key})"
        )
        if self.cfg[FILE_KEY]:
            res += f"\nMultiple Pipelines Allowed: {self.cfg[MULTI_PIPELINE]}"
        else:
            res += f"\nProject Name: {self.cfg[PROJECT_NAME]}"
            res += f"\nDatabase URL: {self.cfg[DB_URL]}"
            res += f"\nConfig File: {self.config_path}"

        res += f"\nPipeline name: {self.cfg[PIPELINE_NAME]}"
        res += f"\nPipeline type: {self.cfg[PIPELINE_TYPE]}"
        if self.cfg[SCHEMA_PATH] is not None:
            res += "\nProject Level Data:"
            for k, v in self.cfg[SCHEMA_KEY].project_level_data.items():
                res += f"\n {k} : {v}"
            res += "\nSample Level Data:"
            for k, v in self.cfg[SCHEMA_KEY].sample_level_data.items():
                res += f"\n {k} : {v}"
        res += f"\nStatus Schema key: {self.cfg[STATUS_SCHEMA_KEY]}"
        res += f"\nResults formatter: {str(self.cfg[RESULT_FORMATTER].__name__)}"
        res += f"\nResults schema source: {self.cfg[SCHEMA_PATH]}"
        res += f"\nStatus schema source: {self.cfg[STATUS_SCHEMA_SOURCE_KEY]}"
        res += f"\nRecords count: {self.record_count}"
        if self.cfg[SCHEMA_PATH] is not None:
            high_res = self.highlighted_results
        else:
            high_res = None
        if high_res:
            res += f"\nHighlighted results: {', '.join(high_res)}"
        return res

    def __getitem__(self, key):
        # This is a wrapper for the retrieve function:
        result = self.retrieve_one(record_identifier=key)
        return result

    def __setitem__(self, key, value):
        # This is a wrapper for the report function:
        result = self.report(record_identifier=key, values=value)
        return result

    def __delitem__(self, key):
        # This is a wrapper for the remove function; it removes the entire record:
        result = self.remove(record_identifier=key)
        return result

    def __iter__(
        self,
        limit: Optional[int] = 1000,
        cursor: Optional[int] = None,
    ) -> Iterator:
        """
        This is a wrapper around select_records that creates an iterator of records

        :param int limit: maximum number of results to retrieve per page
        :param int cursor: cursor position to begin retrieving records
        :return: Iterator
        """
        if self.file:
            # File backend does not support cursor-based paging
            return iter(self.select_records(limit=limit)["records"])
        else:
            return iter(self.select_records(limit=limit, cursor=cursor)["records"])

    def __len__(self):
        return len(self.cfg)

    def resolve_results_file_path(self, results_file_path):
        """Replace {record_identifier} in results_file_path if it exists.
        :param str results_file_path: YAML file to report into, if file is
        used as the object back-end
        """
        # Save for later when assessing if there may be multiple result files
        if results_file_path:
            assert isinstance(results_file_path, str), TypeError("Path is expected to be a str")
            if not self.record_identifier and "{record_identifier}" in results_file_path:
                raise NotImplementedError(
                    f"Must provide record identifier during PipestatManager creation for this results_file_path: {results_file_path}"
                )
            self.cfg["unresolved_result_path"] = results_file_path
            return results_file_path.format(record_identifier=self.record_identifier)
        return results_file_path

    def initialize_filebackend(self, record_identifier, results_file_path, flag_file_dir):
        # Check if there will be multiple results_file_paths
        _LOGGER.debug(f"Determined file as backend: {results_file_path}")

        if self.cfg[DB_ONLY_KEY]:
            _LOGGER.debug(
                "Running in database only mode does not make sense with a YAML file as a backend. "
                "Changing back to using memory."
            )
            self.cfg[DB_ONLY_KEY] = False

        flag_file_dir = self.cfg[CONFIG_KEY].priority_get(
            "flag_file_dir",
            override=flag_file_dir,
            default=os.path.dirname(self.cfg[FILE_KEY]),
        )
        self.cfg[STATUS_FILE_DIR] = mk_abs_via_cfg(
            flag_file_dir, self.config_path or self.cfg[FILE_KEY]
        )

        self.backend = FileBackend(
            self.cfg[FILE_KEY],
            record_identifier,
            self.cfg[PIPELINE_NAME],
            self.cfg[PIPELINE_TYPE],
            self.cfg[SCHEMA_KEY],
            self.cfg[STATUS_SCHEMA_KEY],
            self.cfg[STATUS_FILE_DIR],
            self.cfg[RESULT_FORMATTER],
            self.cfg[MULTI_PIPELINE],
        )

        return

    @check_dependencies(
        dependency_list=["DBBackend"],
        msg="Missing required dependencies for this usage, e.g. try pip install pipestat['dbbackend']",
    )
    def initialize_dbbackend(self, record_identifier, show_db_logs):
        _LOGGER.debug("Determined database as backend")
        if self.cfg[SCHEMA_KEY] is None:
            raise SchemaNotFoundError("Output schema must be supplied for DB backends.")
        if CFG_DATABASE_KEY not in self.cfg[CONFIG_KEY]:
            raise NoBackendSpecifiedError()
        try:
            dbconf = self.cfg[CONFIG_KEY].exp[
                CFG_DATABASE_KEY
            ]  # the .exp expands the paths before url construction
            self.cfg[DB_URL] = construct_db_url(dbconf)
        except KeyError:
            raise PipestatDatabaseError(f"No database section ('{CFG_DATABASE_KEY}') in config")
        self._show_db_logs = show_db_logs

        self.backend = DBBackend(
            record_identifier,
            self.cfg[PIPELINE_NAME],
            show_db_logs,
            self.cfg[PIPELINE_TYPE],
            self.cfg[SCHEMA_KEY],
            self.cfg[STATUS_SCHEMA_KEY],
            self.cfg[DB_URL],
            self.cfg[STATUS_SCHEMA_SOURCE_KEY],
            self.cfg[RESULT_FORMATTER],
        )

    @require_backend
    def clear_status(
        self,
        record_identifier: str = None,
        flag_names: List[str] = None,
    ) -> List[Union[str, None]]:
        """
        Remove status flags

        :param str record_identifier: name of the sample_level record to remove flags for
        :param Iterable[str] flag_names: Names of flags to remove, optional; if
            unspecified, all schema-defined flag names will be used.
        :return List[Union[str,None]]: Collection of names of flags removed
        """

        r_id = record_identifier or self.record_identifier
        return self.backend.clear_status(record_identifier=r_id, flag_names=flag_names)

    @require_backend
    def count_records(self) -> int:
        """
        Count records
        :return int: number of records
        """
        return self.backend.count_records()

    @require_backend
    def get_status(
        self,
        record_identifier: str = None,
    ) -> Optional[str]:
        """
        Get the current pipeline status
        :param str record_identifier: name of the sample_level record
        :return str: status identifier, e.g. 'running'

        """

        r_id = record_identifier or self.record_identifier
        return self.backend.get_status(record_identifier=r_id)

    @require_backend
    def list_recent_results(
        self,
        limit: Optional[int] = 1000,
        start: Optional[str] = None,
        end: Optional[str] = None,
        time_column: Optional[str] = "modified",
    ) -> dict:
        """
        :param int  limit: limit number of results returned
        :param str start: most recent result to filter on, defaults to now, e.g. 2023-10-16 13:03:04.680400
        :param str end: oldest result to filter on, e.g. 1970-10-16 13:03:04.680400
        :param str time_column: created or modified column/attribute to filter on
        :return dict results: a dict containing start, end, num of records, and list of retrieved records
        """

        date_format = "%Y-%m-%d %H:%M:%S"
        if start is None:
            start = datetime.datetime.now()
        else:
            try:
                start = datetime.datetime.strptime(start, date_format)
            except ValueError:
                raise InvalidTimeFormatError(msg=f"Incorrect time format, requires:{date_format}")

        if end is None:
            end = datetime.datetime.strptime("1900-01-01 00:00:00", date_format)
        else:
            try:
                end = datetime.datetime.strptime(end, date_format)
            except ValueError:
                raise InvalidTimeFormatError(msg=f"Incorrect time format, requires: {date_format}")

        if time_column == "created":
            col_name = CREATED_TIME
        else:
            col_name = MODIFIED_TIME
        results = self.select_records(
            limit=limit,
            filter_conditions=[
                {
                    "key": col_name,
                    "operator": "lt",
                    "value": start,
                },
                {
                    "key": col_name,
                    "operator": "gt",
                    "value": end,
                },
            ],
        )

        return results

    def process_schema(self, schema_path):
        # Load pipestat schema in two parts: 1) main and 2) status
        self._schema_path = self.cfg[CONFIG_KEY].priority_get(
            "schema_path", env_var=ENV_VARS["schema"], override=schema_path
        )

        if self._schema_path is None:
            _LOGGER.warning("No schema supplied.")
            self.cfg[SCHEMA_KEY] = None
            self.cfg[STATUS_SCHEMA_KEY] = None
            self.cfg[STATUS_SCHEMA_SOURCE_KEY] = None
            # return None
            # raise SchemaNotFoundError("PipestatManager creation failed; no schema")
        else:
            # Main schema
            schema_to_read = mk_abs_via_cfg(self._schema_path, self.cfg["config_path"])
            self._schema_path = schema_to_read
            parsed_schema = ParsedSchema(schema_to_read)
            self.cfg[SCHEMA_KEY] = parsed_schema

            # Status schema
            self.cfg[STATUS_SCHEMA_KEY] = parsed_schema.status_data
            if not self.cfg[STATUS_SCHEMA_KEY]:
                (
                    self.cfg[STATUS_SCHEMA_SOURCE_KEY],
                    self.cfg[STATUS_SCHEMA_KEY],
                ) = read_yaml_data(path=STATUS_SCHEMA, what="default status schema")
            else:
                self.cfg[STATUS_SCHEMA_SOURCE_KEY] = schema_to_read

    @require_backend
    def remove(
        self,
        record_identifier: str = None,
        result_identifier: str = None,
    ) -> bool:
        """
        Remove a result.

        If no result ID specified or last result is removed, the entire record
        will be removed.

        :param str record_identifier: name of the sample_level record
        :param str result_identifier: name of the result to be removed or None
             if the record should be removed.
        :return bool: whether the result has been removed
        """

        r_id = record_identifier or self.cfg[RECORD_IDENTIFIER]
        return self.backend.remove(
            record_identifier=r_id,
            result_identifier=result_identifier,
        )

    @require_backend
    def report(
        self,
        values: Dict[str, Any],
        record_identifier: Optional[str] = None,
        force_overwrite: bool = False,
        result_formatter: Optional[staticmethod] = None,
        strict_type: bool = True,
    ) -> Union[List[str], bool]:
        """
        Report a result.

        :param Dict[str, any] values: dictionary of result-value pairs
        :param str record_identifier: unique identifier of the record, value
            in 'record_identifier' column to look for to determine if the record
            already exists
        :param bool force_overwrite: whether to overwrite the existing record
        :param str result_formatter: function for formatting result
        :param bool strict_type: whether the type of the reported values should
            remain as is. Pipestat would attempt to convert to the
            schema-defined one otherwise
        :return str reported_results: return list of formatted string
        """

        result_formatter = result_formatter or self.cfg[RESULT_FORMATTER]
        values = deepcopy(values)
        r_id = record_identifier or self.cfg[RECORD_IDENTIFIER]
        if r_id is None:
            raise NotImplementedError("You must supply a record identifier to report results")

        result_identifiers = list(values.keys())
        if self.cfg[SCHEMA_KEY] is not None:
            for r in result_identifiers:
                validate_type(
                    value=values[r],
                    schema=self.result_schemas[r],
                    strict_type=strict_type,
                )

        reported_results = self.backend.report(
            values=values,
            record_identifier=r_id,
            force_overwrite=force_overwrite,
            result_formatter=result_formatter,
        )

        return reported_results

    @require_backend
    def select_distinct(
        self,
        columns: Optional[Union[str, List[str]]] = None,
    ) -> List[Any]:
        """
        Retrieves unique results for a list of attributes.

        :param List[str] columns: columns to include in the result
        :return list[any] result: this is a list of distinct results
        """
        if not isinstance(columns, list) and not isinstance(columns, str):
            raise ValueError(
                "Columns must be a list of strings or string, e.g. ['record_identifier', 'number_of_things']"
            )

        result = self.backend.select_distinct(columns=columns)
        return result

    @require_backend
    def select_records(
        self,
        columns: Optional[List[str]] = None,
        filter_conditions: Optional[List[Dict[str, Any]]] = None,
        limit: Optional[int] = 1000,
        cursor: Optional[int] = None,
        bool_operator: Optional[str] = "AND",
    ) -> Dict[str, Any]:
        """
        :param list[str] columns: columns to include in the result
        :param list[dict]  filter_conditions: e.g. [{"key": ["id"], "operator": "eq", "value": 1)], operator list:
            - eq for ==
            - lt for <
            - ge for >=
            - in for in_
            - like for like
        :param int limit: maximum number of results to retrieve per page
        :param int cursor: cursor position to begin retrieving records
        :param bool bool_operator: Perform filtering with AND or OR Logic.
        :return dict records_dict = {
            "total_size": int,
            "page_size": int,
            "next_page_token": int,
            "records": List[Dict[{key, Any}]],
        }
        """

        return self.backend.select_records(
            columns=columns,
            filter_conditions=filter_conditions,
            limit=limit,
            cursor=cursor,
            bool_operator=bool_operator,
        )

    @require_backend
    def retrieve_one(
        self,
        record_identifier: str = None,
        result_identifier: Optional[Union[str, List[str]]] = None,
    ) -> Union[Any, Dict[str, Any]]:
        """
        Retrieve a single record
        :param str record_identifier: single record_identifier
        :param str result_identifier: single record_identifier
        :return: Dict[str, any]: a mapping with filtered


            results reported for the record
        """
        r_id = record_identifier or self.record_identifier

        filter_conditions = [
            {
                "key": "record_identifier",
                "operator": "eq",
                "value": record_identifier,
            },
        ]
        if result_identifier:
            if isinstance(result_identifier, str):
                columns = [result_identifier]
            elif isinstance(result_identifier, list):
                columns = result_identifier
            else:
                raise ValueError("Result identifier must be a str or list[str]")
            result = self.select_records(filter_conditions=filter_conditions, columns=columns)[
                "records"
            ]
            if len(result) > 0:
                if len(columns) > 1:
                    try:
                        return result[0]
                    except IndexError:
                        raise RecordNotFoundError(
                            f"Results '{columns}' for '{record_identifier}' not found"
                        )
                try:
                    return result[0][result_identifier]
                except IndexError:
                    raise RecordNotFoundError(
                        f"Results '{columns}' for '{record_identifier}' not found"
                    )
            else:
                raise RecordNotFoundError(
                    f"Results '{columns}' for '{record_identifier}' not found"
                )
        else:
            try:
                result = self.select_records(filter_conditions=filter_conditions)["records"]
                if len(result) > 0:
                    try:
                        return result[0]
                    except IndexError:
                        raise RecordNotFoundError(f"Record '{record_identifier}' not found")
                else:
                    raise RecordNotFoundError(f"Record '{record_identifier}' not found")
            except IndexError:
                raise RecordNotFoundError(f"Record '{record_identifier}' not found")

    def retrieve_many(
        self,
        record_identifiers: List[str],
        result_identifier: Optional[str] = None,
    ) -> Union[Any, Dict[str, Any]]:
        """
        :param record_identifiers: list of record identifiers
        :param str result_identifier: single record_identifier
        :return: Dict[str, any]: a mapping with filtered
            results reported for the record
        """

        filter = {
            "key": "record_identifier",
            "operator": "in",
            "value": record_identifiers,
        }
        if result_identifier:
            result = self.select_records(filter_conditions=[filter], columns=[result_identifier])
        else:
            result = self.select_records(filter_conditions=[filter])

        if len(result["records"]) == 0:
            RecordNotFoundError(f"Records, '{record_identifiers}',  not found")
        else:
            return result

    @require_backend
    def set_status(
        self,
        status_identifier: str,
        record_identifier: str = None,
    ) -> None:
        """
        Set pipeline run status.

        The status identifier needs to match one of identifiers specified in
        the status schema. A basic, ready to use, status schema is shipped with
        this package.

        :param str status_identifier: status to set, one of statuses defined
            in the status schema
        :param str record_identifier: sample_level record identifier to set the
            pipeline status for
        """
        r_id = record_identifier or self.record_identifier
        self.backend.set_status(status_identifier, r_id)

    @require_backend
    def link(self, link_dir) -> str:
        """
        This function creates a link structure such that results are organized by type.
        :param str link_dir: path to desired symlink output directory
        :return str linked_results_path: path to symlink directory
        """
        self.check_multi_results()
        linked_results_path = self.backend.link(link_dir=link_dir)

        return linked_results_path

    @require_backend
    def summarize(
        self,
        looper_samples: Optional[list] = None,
        amendment: Optional[str] = None,
        portable: Optional[bool] = False,
    ) -> None:
        """
        Builds a browsable html report for reported results.
        :param Iterable[str] looper_samples: list of looper Samples from PEP
        :param Iterable[str] amendment: name indicating amendment to use, optional
        :param bool portable: moves figures and report files to directory for easy sharing
        :return str: report_path

        """

        self.check_multi_results()

        html_report_builder = HTMLReportBuilder(prj=self, portable=portable)
        report_path = html_report_builder(
            pipeline_name=self.cfg[PIPELINE_NAME],
            amendment=amendment,
            looper_samples=looper_samples,
        )

        if portable is True:
            zip_report(report_dir_name=os.path.dirname(report_path))

        return report_path

    def check_multi_results(self):
        # Check to see if the user used a path with "{record-identifier}"
        if self.file and self.cfg["unresolved_result_path"] != self.file:
            if "{record_identifier}" in self.cfg["unresolved_result_path"]:
                # assume there are multiple result files in sub-directories
                self.cfg["multi_result_files"] = True
                results_directory = self.cfg["unresolved_result_path"].split(
                    "{record_identifier}"
                )[0]
                results_directory = mk_abs_via_cfg(results_directory, self.cfg["config_path"])
                self.backend.aggregate_multi_results(results_directory)

    @require_backend
    def table(
        self,
    ) -> List[str]:
        """
        Generates stats (.tsv) and object (.yaml) files.
        :return list[str] table_path_list: list containing output file paths of stats and objects

        """

        self.check_multi_results()
        pipeline_name = self.cfg[PIPELINE_NAME]
        table_path_list = _create_stats_objs_summaries(self, pipeline_name)

        return table_path_list

    def _get_attr(self, attr: str) -> Any:
        """
        Safely get the name of the selected attribute of this object

        :param str attr: attr to select
        :return:
        """
        return self.get(attr)

    @property
    def config_path(self) -> str:
        """
        Config path. None if the config was not provided or if provided
        as a mapping of the config contents

        :return str: path to the provided config
        """
        return self.cfg.get("config_path", None)

    @property
    def data(self) -> YAMLConfigManager:
        """
        Data object

        :return yacman.YAMLConfigManager: the object that stores the reported data
        """
        return self.backend._data

    @property
    def db_url(self) -> str:
        """
        Database URL, generated based on config credentials

        :return str: database URL
        :raise PipestatDatabaseError: if the object is not backed by a database
        """
        return self.cfg[DB_URL]

    @property
    def file(self) -> str:
        """
        File path that the object is reporting the results into

        :return str: file path that the object is reporting the results into
        """
        return self.cfg[FILE_KEY]

    @property
    def highlighted_results(self) -> List[str]:
        """
        Highlighted results

        :return List[str]: a collection of highlighted results
        """
        return [k for k, v in self.result_schemas.items() if v.get("highlight") is True]

    @property
    def output_dir(self) -> str:
        """
        Output directory for report and stats generation

        :return str: path to output_dir
        """
        return self.cfg[OUTPUT_DIR]

    @property
    def pipeline_name(self) -> str:
        """
        Pipeline name

        :return str: Pipeline name
        """
        return self.cfg[PIPELINE_NAME]

    @property
    def project_name(self) -> str:
        """
        Project name the object writes the results to

        :return str: project name the object writes the results to
        """
        return self.cfg[PROJECT_NAME]

    @property
    def pipeline_type(self) -> str:
        """
        Pipeline type: "sample" or "project"

        :return str: pipeline type
        """
        return self.cfg[PIPELINE_TYPE]

    @property
    def record_identifier(self) -> str:
        """
        Pipeline type: "sample" or "project"

        :return str: pipeline type
        """
        return self.cfg[RECORD_IDENTIFIER]

    @property
    def record_count(self) -> int:
        """
        Number of records reported

        :return int: number of records reported
        """
        return self.count_records()

    @property
    def result_schemas(self) -> Dict[str, Any]:
        """
        Result schema mappings

        :return dict: schemas that formalize the structure of each result
            in a canonical jsonschema way
        """
        return {
            **self.cfg[SCHEMA_KEY].project_level_data,
            **self.cfg[SCHEMA_KEY].sample_level_data,
        }

    @property
    def schema(self) -> ParsedSchema:
        """
        Schema mapping

        :return ParsedSchema: schema object that formalizes the results structure
        """
        return self.cfg["_schema"]

    @property
    def schema_path(self) -> str:
        """
        Schema path

        :return str: path to the provided schema
        """
        return self.cfg[SCHEMA_PATH]

    @property
    def status_schema(self) -> Dict:
        """
        Status schema mapping

        :return dict: schema that formalizes the pipeline status structure
        """
        return self.cfg[STATUS_SCHEMA_KEY]

    @property
    def status_schema_source(self) -> Dict:
        """
        Status schema source

        :return dict: source of the schema that formalizes
            the pipeline status structure
        """
        return self.cfg[STATUS_SCHEMA_SOURCE_KEY]


class SamplePipestatManager(PipestatManager):
    def __init__(self, **kwargs):
        PipestatManager.__init__(self, pipeline_type="sample", **kwargs)
        _LOGGER.warning("Initialize PipestatMgrSample")


class ProjectPipestatManager(PipestatManager):
    def __init__(self, **kwargs):
        PipestatManager.__init__(self, pipeline_type="project", **kwargs)
        _LOGGER.warning("Initialize PipestatMgrProject")


class PipestatBoss(ABC):
    """
    PipestatBoss simply holds Sample or Project Managers that are child classes of PipestatManager.
        :param list[str] pipeline_list: list that holds pipeline types, e.g. ['sample','project']
        :param str record_identifier: record identifier to report for. This
            creates a weak bound to the record, which can be overridden in
            this object method calls
        :param str schema_path: path to the output schema that formalizes
            the results structure
        :param str results_file_path: YAML file to report into, if file is
            used as the object back-end
        :param bool database_only: whether the reported data should not be
            stored in the memory, but only in the database
        :param str | dict config: path to the configuration file or a mapping
            with the config file content
        :param str flag_file_dir: path to directory containing flag files
        :param bool show_db_logs: Defaults to False, toggles showing database logs
        :param str pipeline_type: "sample" or "project"
        :param str result_formatter: function for formatting result
        :param bool multi_pipelines: allows for running multiple pipelines for one file backend
        :param str output_dir: target directory for report generation via summarize and table generation via table.
    """

    def __init__(self, pipeline_list: Optional[list] = None, **kwargs):
        _LOGGER.warning("Initialize PipestatBoss")
        if len(pipeline_list) > 3:
            _LOGGER.warning(
                "PipestatBoss currently only supports one 'sample' and one 'project' pipeline. Ignoring extra types."
            )
        for i in pipeline_list:
            if i == "sample":
                self.samplemanager = SamplePipestatManager(**kwargs)
            elif i == "project":
                self.projectmanager = ProjectPipestatManager(**kwargs)
            else:
                _LOGGER.warning(f"This pipeline type is not supported. Pipeline supplied: {i}")

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)



import datetime
import os.path
import operator
from copy import deepcopy
from functools import reduce
from itertools import chain
from ...helpers import get_all_result_files


from glob import glob
from logging import getLogger
from yacman import YAMLConfigManager
from ubiquerg import create_lock, remove_lock

from typing import List, Dict, Any, Optional, Union, Literal, Callable, Tuple

from ...exceptions import UnrecognizedStatusError, PipestatError
from ...backends.abstract import PipestatBackend
from ...const import DATE_FORMAT, PKG_NAME, CREATED_TIME, MODIFIED_TIME


_LOGGER = getLogger(PKG_NAME)


class FileBackend(PipestatBackend):
    def __init__(
        self,
        results_file_path: str,
        record_identifier: Optional[str] = None,
        pipeline_name: Optional[str] = None,
        pipeline_type: Optional[str] = None,
        parsed_schema: Optional[str] = None,
        status_schema: Optional[str] = None,
        status_file_dir: Optional[str] = None,
        result_formatter: Optional[staticmethod] = None,
        multi_pipelines: Optional[bool] = None,
    ):
        """
        Class representing a File backend
        :param str results_file_path: YAML file to report into, if file is
            used as the object back-end
        :param str record_identifier: record identifier to report for. This
            creates a weak bound to the record, which can be overridden in
            this object method calls
        :param str pipeline_name: name of pipeline associated with result
        :param str pipeline_type: "sample" or "project"
        :param str parsed_schema: results output schema. Used to construct DB columns.
        :param str status_schema: schema containing pipeline statuses e.g. 'running'
        :param str status_file_dir: directory for placing status flags
        :param str result_formatter: function for formatting result
        :param bool multi_pipelines: allows for running multiple pipelines for one file backend

        """
        super().__init__(pipeline_type)
        _LOGGER.warning("Initialize FileBackend")

        self.results_file_path = results_file_path
        self.pipeline_name = pipeline_name
        self.pipeline_type = pipeline_type
        self.record_identifier = record_identifier
        self.parsed_schema = parsed_schema
        self.status_schema = status_schema
        self.status_file_dir = status_file_dir
        self.result_formatter = result_formatter
        self.multi_pipelines = multi_pipelines

        self.determine_results_file(self.results_file_path)

    def determine_results_file(self, results_file_path: str) -> None:
        """Initialize or load results_file from given path
        :param str results_file_path: YAML file to report into, if file is
        used as the object back-end
        """
        if not os.path.exists(self.results_file_path):
            _LOGGER.debug(
                f"Results file doesn't yet exist. Initializing: {self.results_file_path}"
            )
            self._init_results_file()
        else:
            _LOGGER.debug(f"Loading results file: {self.results_file_path}")
            self._load_results_file()

    def check_record_exists(
        self,
        record_identifier: str,
    ) -> bool:
        """
        Check if the specified record exists in self._data

        :param str record_identifier: record to check for
        :return bool: whether the record exists in the table
        """

        return (
            self.pipeline_name in self._data
            and record_identifier in self._data[self.pipeline_name][self.pipeline_type]
        )

    def clear_status(
        self, record_identifier: str = None, flag_names: List[str] = None
    ) -> List[Union[str, None]]:
        """
        Remove status flags

        :param str record_identifier: name of the record to remove flags for
        :param Iterable[str] flag_names: Names of flags to remove, optional; if
            unspecified, all schema-defined flag names will be used.
        :return List[str]: Collection of names of flags removed
        """

        flag_names = flag_names or list(self.status_schema.keys())
        if isinstance(flag_names, str):
            flag_names = [flag_names]
        removed = []
        for f in flag_names:
            path_flag_file = self.get_status_flag_path(
                status_identifier=f, record_identifier=record_identifier
            )
            try:
                os.remove(path_flag_file)
            except:
                pass
            else:
                _LOGGER.info(f"Removed existing flag: {path_flag_file}")
                removed.append(f)
        return removed

    def count_records(self):
        """
        Count records
        :return int: number of records
        """

        return len(self._data[self.pipeline_name])

    def get_flag_file(self, record_identifier: str = None) -> Union[str, List[str], None]:
        """
        Get path to the status flag file for the specified record

        :param str record_identifier: unique record identifier
        :return str | list[str] | None: path to the status flag file
        """

        r_id = record_identifier
        regex = os.path.join(self.status_file_dir, f"{self.pipeline_name}_{r_id}_*.flag")
        file_list = glob(regex)
        if len(file_list) > 1:
            _LOGGER.warning("Multiple flag files found")
            return file_list
        elif len(file_list) == 1:
            return file_list[0]
        else:
            _LOGGER.debug("No flag files found")
            return None
        pass

    def get_status(self, record_identifier: str) -> Optional[str]:
        """
        Get the current pipeline status

        :param str record_identifier: record identifier to set the
            pipeline status for
        :return str: status identifier, e.g. 'running'
        """
        r_id = record_identifier or self.record_identifier
        flag_file = self.get_flag_file(record_identifier=record_identifier)
        if flag_file is not None:
            assert isinstance(flag_file, str), TypeError(
                "Flag file path is expected to be a str, were multiple flags found?"
            )
            with open(flag_file, "r") as f:
                status = f.read()
            return status
        _LOGGER.debug(
            f"Could not determine status for '{r_id}' record. "
            f"No flags found in: {self.status_file_dir}"
        )
        return None

    def get_status_flag_path(self, status_identifier: str, record_identifier=None) -> str:
        """
        Get the path to the status file flag

        :param str status_identifier: one of the defined status IDs in schema
        :param str record_identifier: unique record ID
        :return str: absolute path to the flag file or None if object is
            backed by a DB
        """

        r_id = record_identifier
        return os.path.join(
            self.status_file_dir,
            f"{self.pipeline_name}_{r_id}_{status_identifier}.flag",
        )

    def list_results(
        self,
        restrict_to: Optional[List[str]] = None,
        record_identifier: Optional[str] = None,
    ) -> List[str]:
        """
        Lists all, or a selected set of, reported results

        :param List[str] restrict_to: selected subset of names of results to list
        :param str record_identifier: unique identifier of the record
        :return List[str]: names of results which exist
        """
        record_identifier = record_identifier or self.record_identifier

        try:
            results = list(
                self._data[self.pipeline_name][self.pipeline_type][record_identifier].keys()
            )
        except KeyError:
            return []
        if restrict_to:
            return [r for r in restrict_to if r in results]
        return results

    def remove(
        self,
        record_identifier: Optional[str] = None,
        result_identifier: Optional[str] = None,
    ) -> bool:
        """
        Remove a result.

        If no result ID specified or last result is removed, the entire record
        will be removed.

        :param str record_identifier: unique identifier of the record
        :param str result_identifier: name of the result to be removed or None
             if the record should be removed.
        :return bool: whether the result has been removed
        """

        record_identifier = record_identifier or self.record_identifier

        rm_record = True if result_identifier is None else False

        if not self.check_record_exists(
            record_identifier=record_identifier,
        ):
            _LOGGER.error(f"Record '{record_identifier}' not found")
            return False

        if result_identifier and not self.check_result_exists(
            result_identifier, record_identifier
        ):
            _LOGGER.error(f"'{result_identifier}' has not been reported for '{record_identifier}'")
            return False

        if rm_record:
            self.remove_record(
                record_identifier=record_identifier,
                rm_record=rm_record,
            )
        else:
            del self._data[self.pipeline_name][self.pipeline_type][record_identifier][
                result_identifier
            ]
            _LOGGER.info(
                f"Removed result '{result_identifier}' for record "
                f"'{record_identifier}' from '{self.pipeline_name}' namespace"
            )
            if not self._data[self.pipeline_name][self.pipeline_type][record_identifier]:
                _LOGGER.info(
                    f"Last result removed for '{record_identifier}'. " f"Removing the record"
                )
                rm_record = True
                self.remove_record(
                    record_identifier=record_identifier,
                    rm_record=rm_record,
                )
            # Check if the last remaining attributes are the timestamps
            remaining_attributes = list(
                self._data[self.pipeline_name][self.pipeline_type][record_identifier].keys()
            )
            if (
                len(remaining_attributes) == 2
                and CREATED_TIME in remaining_attributes
                and MODIFIED_TIME in remaining_attributes
            ):
                _LOGGER.info(
                    f"Last result removed for '{record_identifier}'. " f"Removing the record"
                )
                rm_record = True
                self.remove_record(
                    record_identifier=record_identifier,
                    rm_record=rm_record,
                )
            with self._data as locked_data:
                locked_data.write()
        return True

    def remove_record(
        self,
        record_identifier: Optional[str] = None,
        rm_record: Optional[bool] = False,
    ) -> bool:
        """
        Remove a record, requires rm_record to be True

        :param str record_identifier: unique identifier of the record
        :param bool rm_record: bool for removing record.
        :return bool: whether the result has been removed
        """
        if rm_record:
            try:
                _LOGGER.info(f"Removing '{record_identifier}' record")
                del self._data[self.pipeline_name][self.pipeline_type][record_identifier]
                with self._data as locked_data:
                    locked_data.write()
                return True
            except:
                _LOGGER.warning(
                    f" Unable to remove record, aborting Removing '{record_identifier}' record"
                )
                return False
        else:
            _LOGGER.info(f" rm_record flag False, aborting Removing '{record_identifier}' record")

    def report(
        self,
        values: Dict[str, Any],
        record_identifier: Optional[str] = None,
        force_overwrite: bool = False,
        result_formatter: Optional[staticmethod] = None,
    ) -> Union[List[str], bool]:
        """
        Update the value of a result in a current namespace.

        This method overwrites any existing data and creates the required
         hierarchical mapping structure if needed.

        :param Dict[str, Any] values: dict of results identifiers and values
            to be reported
        :param str record_identifier: unique identifier of the record
        :param bool force_overwrite: Toggles force overwriting results, defaults to False
        :param str result_formatter: function for formatting result
        :return bool | list[str] results_formatted: return list of formatted string
        """

        # record_identifier = record_identifier or self.record_identifier
        record_identifier = record_identifier

        result_formatter = result_formatter or self.result_formatter
        results_formatted = []
        current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        result_identifiers = list(values.keys())
        if self.parsed_schema is not None:
            self.assert_results_defined(
                results=result_identifiers, pipeline_type=self.pipeline_type
            )
        existing = self.list_results(
            record_identifier=record_identifier,
            restrict_to=result_identifiers,
        )
        if existing:
            existing_str = ", ".join(existing)
            _LOGGER.warning(f"These results exist for '{record_identifier}': {existing_str}")
            if not force_overwrite:
                return False
            _LOGGER.info(f"Overwriting existing results: {existing_str}")
            values.update({MODIFIED_TIME: current_time})
        if not existing:
            if record_identifier in self._data[self.pipeline_name][self.pipeline_type].keys():
                values.update({MODIFIED_TIME: current_time})
            else:
                values.update({CREATED_TIME: current_time})
                values.update({MODIFIED_TIME: current_time})

        self._data[self.pipeline_name][self.pipeline_type].setdefault(record_identifier, {})

        for res_id, val in values.items():
            self._data[self.pipeline_name][self.pipeline_type][record_identifier][res_id] = val
            results_formatted.append(
                result_formatter(
                    pipeline_name=self.pipeline_name,
                    record_identifier=record_identifier,
                    res_id=res_id,
                    value=val,
                )
            )

        with self._data as locked_data:
            locked_data.write()

        return results_formatted

    def select_distinct(self, columns: Union[str, List[str]]) -> List[Tuple]:
        """
        Retrieve distinct results given a list of columns

        :param str | List[str] columns: column columns to include in the result
        :return List[Tuple]: returns distinct values.
        """
        if isinstance(columns, str):
            columns = [columns]

        record_data = self._data.data[self.pipeline_name][self.pipeline_type]

        final_values_list = []
        for record_id in record_data.keys():
            record_value_list = []
            for column in columns:
                if column in list(record_data[record_id].keys()):
                    record_value_list.append(record_data[record_id][column])
                else:
                    if column != "record_identifier":
                        _LOGGER.warning(msg=f"Column {column} not reported, skipping....")

            if "record_identifier" in columns:
                # record_identifier is not a column in the file backend but the user may want it
                record_value_list.append(record_id)
            final_values_list.append(tuple(record_value_list))

        # make sure that the order is correct
        unique_lists = list(set(final_values_list))
        return unique_lists

    def select_records(
        self,
        columns: Optional[List[str]] = None,
        filter_conditions: Optional[List[Dict[str, Any]]] = None,
        limit: Optional[int] = 1000,
        cursor: Optional[int] = None,
        bool_operator: Optional[str] = "AND",
    ) -> Dict[str, Any]:
        """
        Select records from the FileBackend

        :param list[str] columns: columns to include in the result
        :param list[dict]  filter_conditions: e.g. [{"key": ["id"], "operator": "eq", "value": 1)], operator list:
            - eq for ==
            - lt for <
            - ge for >=
            - in for in_
        :param int limit: maximum number of results to retrieve per page
        :param int cursor: cursor position to begin retrieving records
        :param bool bool_operator: Perform filtering with AND or OR Logic.
        :return dict records_dict = {
            "total_size": int,
            "page_size": int,
            "next_page_token": int,
            "records": List[Dict[{key, Any}]],
        }
        """
        if cursor:
            _LOGGER.warning("Cursor not supported for FileBackend, ignoring cursor")

        def get_operator(op: Literal["eq", "lt", "ge", "gt", "in"]) -> Any:
            """
            Get python operator for a given string

            :param str op: desired operator, "eq", "lt"
            :return: operator function
            """

            if op == "eq":
                return operator.__eq__
            if op == "lt":
                return operator.__lt__
            if op == "ge":
                return operator.__ge__
            if op == "gt":
                return operator.__gt__
            if op == "in":
                return operator.contains
            raise ValueError(f"Invalid filter operator: {op}")

        def get_nested_column(result_value: dict, key_list: list, retrieved_operator: Callable):
            """
            Recursive function that evaluates a nested list of keys vs a value dict

            :param dict result_value: nested dictionary
            :param key_list: list of keys, e.g. keys that may be in the nested dictionary e.g. ['id', 'name']
            :param retrieved_operator: operator function (ge, gt...)
            :return bool:
            """
            if len(key_list) == 1:
                if result_value.get(key_list[0], None):
                    if retrieved_operator(key_list, result_value.get(key_list[0])):
                        return True
                return False
            else:
                return get_nested_column(
                    result_value[key_list[0]], key_list[1:], retrieved_operator
                )

        records_list = []

        data = deepcopy(self._data.data[self.pipeline_name][self.pipeline_type])

        if columns:
            if not isinstance(columns, list):
                raise ValueError(
                    "Columns must be a list of strings, e.g. ['record_identifier', 'number_of_things']"
                )

        total_count = len(data.keys())

        filtered_records_list = []
        if filter_conditions:
            for filter_condition in filter_conditions:
                if list(filter_condition.keys()) != ["key", "operator", "value"]:
                    raise ValueError(
                        "Filter conditions must be a dictionary with keys 'key', 'operator', and 'value'"
                    )

                retrieved_operator = get_operator(filter_condition["operator"])
                retrieved_results = []

                # Check each sample's dict
                for record_identifier in list(data.keys())[0:limit]:
                    if filter_condition["key"] != "record_identifier":
                        for key, value in data[record_identifier].items():
                            result = False
                            if isinstance(value, dict):
                                if key == filter_condition["key"][0]:
                                    result = get_nested_column(
                                        value,
                                        filter_condition["key"][1:],
                                        retrieved_operator,
                                    )
                            else:
                                if filter_condition["key"] == key:
                                    # Filter datetime objects
                                    if key in CREATED_TIME or key in MODIFIED_TIME:
                                        try:
                                            time_stamp = datetime.datetime.strptime(
                                                data[record_identifier][key],
                                                DATE_FORMAT,
                                            )
                                            result = retrieved_operator(
                                                time_stamp, filter_condition["value"]
                                            )
                                        except TypeError:
                                            result = False
                                    else:
                                        result = retrieved_operator(
                                            value, filter_condition["value"]
                                        )

                            if result:
                                retrieved_results.append(record_identifier)
                    else:
                        # If user wants record_identifier
                        if record_identifier in filter_condition["value"]:
                            retrieved_results.append(record_identifier)

                if retrieved_results:
                    filtered_records_list.append(retrieved_results)
        else:
            # Assume user wants all the records if no filter was given.
            filtered_records_list = [list(data.keys())[0:limit]]

        # There is now a list of dicts for each filtered condition.
        # Depending on Union or Intersection we want to pare down the list.
        shared_keys = []

        if bool_operator.lower() == "and" and filtered_records_list:
            shared_keys = list(reduce(lambda i, j: i & j, (set(x) for x in filtered_records_list)))

        if bool_operator.lower() == "or" and filtered_records_list:
            shared_keys = list(set(chain(*filtered_records_list)))

        if shared_keys:
            for record_identifier in sorted(shared_keys):
                record = {}
                if columns:  # Did the user specify a list of columns as well?
                    for key, value in list(data[record_identifier].items()):
                        if key in columns:
                            record.update({key: value})
                    if "record_identifier" in columns:
                        record.update({"record_identifier": record_identifier})
                else:
                    record = data[record_identifier]
                if record != {}:
                    record.update({"record_identifier": record_identifier})
                    records_list.append(record)

        records_dict = {
            "total_size": total_count,
            "page_size": limit,
            "next_page_token": 0,
            "records": records_list,
        }

        return records_dict

    def set_status(
        self,
        status_identifier: str,
        record_identifier: str = None,
    ) -> None:
        """
        Set pipeline run status.

        The status identifier needs to match one of identifiers specified in
        the status schema. A basic, ready to use, status schema is shipped with
        this package.

        :param str status_identifier: status to set, one of statuses defined
            in the status schema
        :param str record_identifier: record identifier to set the
            pipeline status for
        """
        r_id = record_identifier or self.record_identifier
        if self.status_schema is not None:
            known_status_identifiers = self.status_schema.keys()
            if status_identifier not in known_status_identifiers:
                raise UnrecognizedStatusError(
                    f"'{status_identifier}' is not a defined status identifier. "
                    f"These are allowed: {known_status_identifiers}"
                )
        prev_status = self.get_status(r_id)

        if prev_status is not None:
            prev_flag_path = self.get_status_flag_path(prev_status, record_identifier)
            os.remove(prev_flag_path)
        flag_path = self.get_status_flag_path(status_identifier, record_identifier)
        create_lock(flag_path)
        with open(flag_path, "w") as f:
            f.write(status_identifier)
        remove_lock(flag_path)

        if prev_status:
            _LOGGER.debug(f"Changed status from '{prev_status}' to '{status_identifier}'")

    def _htmlreportbuilder(self):
        """
        build html report based on all reported results
        """

        # build new folder for the report
        self.reports_dir = os.path.join(self.results_file_path, "reports")
        _LOGGER.debug(f"Reports dir: {self.reports_dir}")

    def _init_results_file(self) -> None:
        """
        Initialize YAML results file if it does not exist.
        Read the data stored in the existing file into the memory otherwise.

        :return bool: whether the file has been created
        """
        _LOGGER.info(f"Initializing results file '{self.results_file_path}'")

        # Must ensure sub-directories exist if they do not
        # TODO should this actually be handled by yacman?
        try:
            os.makedirs(os.path.dirname(self.results_file_path))
        except FileExistsError:
            pass

        self._data = YAMLConfigManager(
            entries={self.pipeline_name: {}},
            filepath=self.results_file_path,
            create_file=True,
        )
        self._data.setdefault(self.pipeline_name, {})
        self._data[self.pipeline_name].setdefault("project", {})
        self._data[self.pipeline_name].setdefault("sample", {})
        with self._data as data_locked:
            data_locked.write()

    def aggregate_multi_results(self, results_directory) -> None:
        """
        Collects single results files and aggregates them into a new aggregate_results.yaml file
        :param str results_directory: directory containing subdirectories containing results.yaml files
        :return: None
        """
        all_result_files = get_all_result_files(results_directory)
        if len(all_result_files) == 0:
            _LOGGER.warning(
                "Attempting to aggregate multiple results files but no result files found. Ensure they are in subdirectories, e.g. 'record1/record1_results.yaml'"
            )
        aggregate_results_file_path = os.path.join(results_directory, "aggregate_results.yaml")

        # THIS WILL OVERWRITE self.results_file_path and self._data on the current psm!
        self.results_file_path = aggregate_results_file_path
        self._init_results_file()

        for file in all_result_files:
            try:
                temp_data = YAMLConfigManager(filepath=file)
            except ValueError:
                temp_data = YAMLConfigManager()
            if self.pipeline_name in temp_data:
                if "project" in temp_data[self.pipeline_name]:
                    self._data[self.pipeline_name]["project"].update(
                        temp_data[self.pipeline_name]["project"]
                    )
                if "sample" in temp_data[self.pipeline_name]:
                    self._data[self.pipeline_name]["sample"].update(
                        temp_data[self.pipeline_name]["sample"]
                    )

        with self._data as data_locked:
            data_locked.write()

    def _load_results_file(self) -> None:
        _LOGGER.debug(f"Reading data from '{self.results_file_path}'")
        data = YAMLConfigManager(filepath=self.results_file_path)
        if not bool(data):
            self._data = data
            self._data.setdefault(self.pipeline_name, {})
            self._data[self.pipeline_name].setdefault("project", {})
            self._data[self.pipeline_name].setdefault("sample", {})
            with self._data as data_locked:
                data_locked.write()
        namespaces_reported = [k for k in data.keys() if not k.startswith("_")]
        num_namespaces = len(namespaces_reported)
        if num_namespaces == 0:
            self._data = data
        elif num_namespaces > 0:
            if self.pipeline_name in namespaces_reported:
                self._data = data
            elif self.pipeline_name not in namespaces_reported and self.multi_pipelines is True:
                self._data = data
                self._data.setdefault(self.pipeline_name, {})
                self._data[self.pipeline_name].setdefault("project", {})
                self._data[self.pipeline_name].setdefault("sample", {})
                _LOGGER.warning("MULTI PIPELINES FOR SINGLE RESULTS FILE")
            else:
                raise PipestatError(
                    f"'{self.results_file_path}' is already in use for {num_namespaces} namespaces: {', '.join(namespaces_reported)} and multi_pipelines = False."
                )



""" Package exception types """

from typing import Iterable, Optional
from .const import CLASSES_BY_TYPE, ENV_VARS

__all__ = [
    "InvalidTypeError",
    "IncompatibleClassError",
    "NoBackendSpecifiedError",
    "PipestatError",
    "PipestatDatabaseError",
    "MissingConfigDataError",
    "SchemaError",
    "SchemaNotFoundError",
    "PipestatDataError",
    "UnrecognizedStatusError",
    "RecordNotFoundError",
    "PipelineTypeNotSuppliedError",
    "InvalidTimeFormatError",
    "PipestatDependencyError",
    "ColumnNotFoundError",
]


class RecordNotFoundError(LookupError):
    def __init__(self, msg):
        super(RecordNotFoundError, self).__init__(msg)


class ColumnNotFoundError(LookupError):
    def __init__(self, msg):
        super(ColumnNotFoundError, self).__init__(msg)


class PipelineTypeNotSuppliedError(LookupError):
    def __init__(self, msg):
        super(PipelineTypeNotSuppliedError, self).__init__(msg)


class PipestatError(Exception):
    """Base exception type for this package"""


class NoBackendSpecifiedError(PipestatError):
    """Subtype for designating lack of backend specification"""


class SchemaError(PipestatError):
    """Schema error"""

    def __init__(self, msg):
        super(SchemaError, self).__init__(msg)


class SchemaNotFoundError(SchemaError):
    """Schema not found error"""

    def __init__(self, msg, cli=False):
        txt = f"Results schema not found. The schema is required to {msg}. "
        txt += (
            "It needs to be supplied as an CLI argument"
            if cli
            else "It needs to be supplied to the object constructor"
        )
        txt += f" or via '{ENV_VARS['schema']}' environment variable."
        super(SchemaNotFoundError, self).__init__(txt)


class MissingConfigDataError(PipestatError):
    """Exception for invalid config file."""

    def __init__(self, msg):
        spacing = " " if msg[-1] in ["?", ".", "\n"] else "; "
        suggest = "For config format documentation please see: http://pipestat.databio.org/en/latest/db_config/"
        super(MissingConfigDataError, self).__init__(msg + spacing + suggest)


class PipestatStartupError(PipestatError):
    """Data error for local data associated with file backend"""

    def __init__(self, msg):
        super(PipestatStartupError, self).__init__(msg)


class PipestatDataError(PipestatError):
    """Data error for local data associated with file backend"""

    def __init__(self, msg):
        super(PipestatDataError, self).__init__(msg)


class InvalidTimeFormatError(PipestatError):
    """Data error for local data associated with file backend"""

    def __init__(self, msg):
        super(InvalidTimeFormatError, self).__init__(msg)


class PipestatDependencyError(PipestatError):
    """Dependency error"""

    def __init__(self, msg):
        super(PipestatDependencyError, self).__init__(msg)


class PipestatDatabaseError(PipestatError):
    """Database error"""

    def __init__(self, msg):
        super(PipestatDatabaseError, self).__init__(msg)


class InvalidTypeError(PipestatError):
    """Type of the reported value is not supported"""

    def __init__(self, type):
        super(InvalidTypeError, self).__init__(
            "'{}' is an invalid type. Only the following types are "
            "supported: {}".format(type, list(CLASSES_BY_TYPE.keys()))
        )


class IncompatibleClassError(PipestatError):
    """Class  of the reported value is not supported"""

    def __init__(self, cls, req_cls, type):
        super(IncompatibleClassError, self).__init__(
            "Incompatible value class for the declared result type ({}). "
            "Required: {}; got: {}".format(type, req_cls, cls)
        )


class UnrecognizedStatusError(PipestatError):
    """Exception for when a value to set as status isn't declared in the active status schema."""

    def __init__(self, status: str, known: Optional[Iterable[str]] = None):
        self._status = status
        msg = f"Unrecognized status: {status}"
        if known is not None:
            pass
        super(UnrecognizedStatusError, self).__init__(msg)

    @property
    def status(self):
        return self._status



import pytest

from pipestat import SamplePipestatManager
from pipestat.const import *
from .conftest import DB_URL

from .conftest import SERVICE_UNAVAILABLE, DB_DEPENDENCIES

try:
    from sqlmodel import SQLModel, create_engine

    from sqlmodel.main import default_registry
except ModuleNotFoundError:
    pass


class ContextManagerDBTesting:
    """
    Creates context manager to connect to database at db_url and drop everything from the database upon exit to ensure
    the db is empty for each new test.
    """

    def __init__(self, db_url):
        self.db_url = db_url

    def __enter__(self):
        self.engine = create_engine(self.db_url, echo=True)
        self.connection = self.engine.connect()
        return self.connection

    def __exit__(self, exc_type, exc_value, exc_traceback):
        SQLModel.metadata.drop_all(self.engine)
        default_registry.dispose()
        self.connection.close()


@pytest.mark.skipif(not DB_DEPENDENCIES, reason="Requires dependencies")
@pytest.mark.skipif(SERVICE_UNAVAILABLE, reason="requires service X to be available")
class TestDatabaseOnly:
    # TODO: parameterize this against different schemas.
    def test_manager_can_be_built_without_exception(self, config_file_path, schema_file_path):
        with ContextManagerDBTesting(DB_URL):
            try:
                SamplePipestatManager(
                    schema_path=schema_file_path,
                    record_identifier="irrelevant",
                    database_only=True,
                    config_file=config_file_path,
                )
            except Exception as e:
                pytest.fail(f"Pipestat manager construction failed: {e})")

    @pytest.mark.parametrize(["rec_id", "res_id"], [("sample2", "number_of_things")])
    def test_select_invalid_filter_column__raises_expected_exception(
        self,
        rec_id,
        res_id,
        config_file_path,
        schema_file_path,
    ):
        with ContextManagerDBTesting(DB_URL):
            args = dict(schema_path=schema_file_path, config_file=config_file_path)
            psm = SamplePipestatManager(**args)
            with pytest.raises(ValueError):
                assert psm.select_records(
                    columns=[res_id],
                    filter_conditions=[
                        {
                            "key": "bogus_name",
                            "operator": "eq",
                            "value": rec_id,
                        },
                    ],
                )

    @pytest.mark.parametrize("res_id", ["number_of_things"])
    @pytest.mark.parametrize("filter_condition", [("column", "eq", 1), "a", [1, 2, 3]])
    def test_select_invalid_filter_structure__raises_expected_exception(
        self,
        res_id,
        config_file_path,
        schema_file_path,
        filter_condition,
    ):
        with ContextManagerDBTesting(DB_URL):
            args = dict(schema_path=schema_file_path, config_file=config_file_path)
            psm = SamplePipestatManager(**args)
            with pytest.raises(AttributeError):
                psm.select_records(
                    filter_conditions=[filter_condition],
                    columns=[res_id],
                )

    @pytest.mark.parametrize(["rec_id", "res_id"], [("sample2", "number_of_things")])
    @pytest.mark.parametrize("limit", [1, 2, 3, 15555])
    def test_select_limit(
        self,
        rec_id,
        res_id,
        config_file_path,
        schema_file_path,
        limit,
    ):
        with ContextManagerDBTesting(DB_URL):
            args = dict(schema_path=schema_file_path, config_file=config_file_path)
            psm = SamplePipestatManager(**args)
            result = psm.select_records(
                filter_conditions=[
                    {
                        "key": "record_identifier",
                        "operator": "eq",
                        "value": rec_id,
                    }
                ],
                columns=[res_id],
                limit=limit,
            )
            assert len(result["records"]) <= limit

    @pytest.mark.parametrize(
        ["offset", "limit"], [(0, 0), (0, 1), (0, 2), (0, 11111), (1, 1), (1, 0)]
    )
    def test_select_pagination(
        self,
        config_file_path,
        schema_file_path,
        offset,
        limit,
    ):
        with ContextManagerDBTesting(DB_URL):
            args = dict(schema_path=schema_file_path, config_file=config_file_path)
            psm = SamplePipestatManager(**args)
            result = psm.select_records(cursor=offset, limit=limit)
            print(result)
            assert len(result["records"]) == min(max((psm.record_count - offset), 0), limit)



"""Assorted project utilities"""

import logging
import glob
import os
import errno

import jsonschema
from json import dumps
from pathlib import Path
from shutil import make_archive
from typing import Any, Dict, Optional, Tuple, Union, List

from oyaml import safe_load, dump
from ubiquerg import expandpath

from zipfile import ZipFile, ZIP_DEFLATED

from .const import (
    PIPESTAT_GENERIC_CONFIG,
    SCHEMA_PROP_KEY,
    SCHEMA_TYPE_KEY,
    CLASSES_BY_TYPE,
)

_LOGGER = logging.getLogger(__name__)


def validate_type(value, schema, strict_type=False):
    """
    Validate reported result against a partial schema, in case of failure try
    to cast the value into the required class.

    Does not support objects of objects.

    :param any value: reported value
    :param dict schema: partial jsonschema schema to validate
        against, e.g. {"type": "integer"}
    :param bool strict_type: whether the value should validate as is
    """

    try:
        jsonschema.validate(value, schema)
    except jsonschema.exceptions.ValidationError as e:
        if strict_type:
            raise
        _LOGGER.debug(f"{str(e)}")
        if schema[SCHEMA_TYPE_KEY] != "object":
            value = CLASSES_BY_TYPE[schema[SCHEMA_TYPE_KEY]](value)
        else:
            for prop, prop_dict in schema[SCHEMA_PROP_KEY].items():
                try:
                    cls_fun = CLASSES_BY_TYPE[prop_dict[SCHEMA_TYPE_KEY]]
                    value[prop] = cls_fun(value[prop])
                except Exception as e:
                    _LOGGER.error(f"Could not cast the result into " f"required type: {str(e)}")
                else:
                    _LOGGER.debug(
                        f"Casted the reported result into required " f"type: {str(cls_fun)}"
                    )
        jsonschema.validate(value, schema)
    else:
        _LOGGER.debug(f"Value '{value}' validated successfully against a schema")


def read_yaml_data(path: Union[str, Path], what: str) -> Tuple[str, Dict[str, Any]]:
    """
    Safely read YAML file and log message

    :param str path: YAML file to read
    :param str what: context
    :return (str, dict): absolute path to the read file and the read data
    """
    if isinstance(path, Path):
        test = lambda p: p.is_file()
    elif isinstance(path, str):
        path = expandpath(path)
        test = os.path.isfile
    else:
        raise TypeError(f"Alleged path to YAML file to read is neither path nor string: {path}")
    assert test(path), FileNotFoundError(f"File not found: {path}")
    _LOGGER.debug(f"Reading {what} from '{path}'")
    with open(path, "r") as f:
        return path, safe_load(f)


def mk_list_of_str(x):
    """
    Make sure the input is a list of strings
    :param str | list[str] | falsy x: input to covert
    :return list[str]: converted input
    :raise TypeError: if the argument cannot be converted
    """
    if not x or isinstance(x, list):
        return x
    if isinstance(x, str):
        return [x]
    raise TypeError(
        f"String or list of strings required as input. Got: " f"{x.__class__.__name__}"
    )


def mk_abs_via_cfg(
    path: Optional[str],
    cfg_path: Optional[str],
) -> Optional[str]:
    """
    Helper function to ensure a path is absolute.

    Assumes a relative path is relative to cfg_path, or to current working directory if cfg_path is None.

    : param str path: The path to make absolute.
    : param str cfg_path: Relative paths will be relative the containing folder of this pat
    """
    if path is None:
        return path
    assert isinstance(path, str), TypeError("Path is expected to be a str")
    if os.path.isabs(path):
        return path
    if cfg_path is None:
        rel_to_cwd = os.path.join(os.getcwd(), path)
        try:
            os.makedirs(os.path.dirname(rel_to_cwd))
        except FileExistsError:
            pass
        if os.path.exists(rel_to_cwd) or os.access(os.path.dirname(rel_to_cwd), os.W_OK):
            return rel_to_cwd
        else:
            raise OSError(f"File not found: {path}")
    joined = os.path.join(os.path.dirname(cfg_path), path)
    try:
        os.makedirs(os.path.dirname(joined))
    except FileExistsError:
        pass
    if os.path.isabs(joined):
        return joined
    raise OSError(f"Could not make this path absolute: {path}")


def init_generic_config():
    """
    Create generic config file for DB Backend
    """
    try:
        os.makedirs("config")
    except FileExistsError:
        pass

    # Destination one level down from CWD in config folder
    dest_file = os.path.join(os.getcwd(), "config", PIPESTAT_GENERIC_CONFIG)

    # Determine Generic Configuration File
    generic_config_dict = {
        "project_name": "generic_test_name",
        "sample_name": "sample1",
        "schema_path": "sample_output_schema.yaml",
        "database": {
            "dialect": "postgresql",
            "driver": "psycopg",
            "name": "pipestat-test",
            "user": "postgres",
            "password": "pipestat-password",
            "host": "127.0.0.1",
            "port": 5432,
        },
    }
    # Write file
    if not os.path.exists(dest_file):
        with open(dest_file, "w") as file:
            dump(generic_config_dict, file)
        print(f"Generic configuration file successfully created at: {dest_file}")
    else:
        print(f"Generic configuration file already exists `{dest_file}`. Skipping creation..")

    return True


def markdown_formatter(pipeline_name, record_identifier, res_id, value) -> str:
    """
    Returns Markdown formatted value as string
    """
    if not isinstance(value, dict):
        nl = "\n"
        rep_strs = [f"`{res_id}`: ```{value}```"]
        formatted_result = (
            f"\n > Reported records for `'{record_identifier}'` in `'{pipeline_name}'` {nl} "
            + f"{nl} {(nl).join(rep_strs)}"
        )
    else:
        nl = "\n"
        rep_strs = [f"`{res_id}`:\n ```\n{dumps(value, indent=2)}\n```"]
        formatted_result = (
            f"\n > Reported records for `'{record_identifier}'` in `'{pipeline_name}'` {nl} "
            + f"{nl} {(nl).join(rep_strs)}"
        )
    return formatted_result


def default_formatter(pipeline_name, record_identifier, res_id, value) -> str:
    """
    Returns formatted value as string
    """
    # Assume default method desired
    nl = "\n"
    rep_strs = [f"{res_id}: {value}"]
    formatted_result = (
        f"Reported records for '{record_identifier}' in '{pipeline_name}' "
        + f":{nl} - {(nl + ' - ').join(rep_strs)}"
    )
    return formatted_result


def force_symlink(file1, file2):
    """Create a symlink between two files."""
    try:
        os.symlink(file1, file2)
    except OSError as e:
        if e.errno == errno.EEXIST:
            _LOGGER.warning(
                f"Symlink collision detected for {file1} and {file2}. Overwriting symlink."
            )
            os.remove(file2)
            os.symlink(file1, file2)


def get_all_result_files(results_file_path: str) -> List:
    """
    Collects any yaml result files relative to the CURRENT results_file_path
    :param str results_file_path: path to the pipestamanager's current result_file
    :return: list
    """
    files = glob.glob(results_file_path + "**/*.yaml")

    return files


def zip_report(report_dir_name: str):
    """

    Walks through files and attempts to zip them into a Zip object using default compression.
    Gracefully fails and informs user if compression library is not available.

    :param report_dir_name: directory name of report directory
    :return: None
    """

    zip_file_name = f"{report_dir_name}_report_portable"

    try:
        make_archive(zip_file_name, "zip", report_dir_name)
    except RuntimeError as e:
        _LOGGER.warning("Report zip file not created! \n {e}")

    if os.path.exists(zip_file_name + ".zip"):
        _LOGGER.info(f"Report zip file successfully created: {zip_file_name}")
    else:
        _LOGGER.warning("Report zip file not created.")



import fastapi
import os
import logging
import uvicorn
from typing import Optional, List, Union, Tuple

from pipestat import SamplePipestatManager
from pipestat.exceptions import RecordNotFoundError
from pipestat.reports import fetch_pipeline_results
from pydantic import BaseModel


_LOGGER = logging.getLogger(__name__)

app = fastapi.FastAPI(
    title="Pipestat Reader",
    description="Allows reading pipestat files from a PostgreSQL Database",
    version="0.01",
)


class FilterQuery(BaseModel):
    column_names: Union[List[str], None] = None
    filter_conditions: Union[List[Tuple[str, str, str]], None] = None


@app.get("/")
async def home():
    """
    Display the home page
    """
    return {"Home": "Welcome"}


@app.get("/data/{record_identifier}")
async def retrieve_results_one_record(record_identifier: str):
    """
    Get all the results for one record
    """
    try:
        result = psm.retrieve(record_identifier=record_identifier)
    except RecordNotFoundError:
        return {"result": "Record not found"}
    return {"result": result}


@app.get("/data/")
async def retrieve_all_records():
    """
    Get all reported records
    """
    try:
        result = psm.get_records()
    except RecordNotFoundError:
        return {"result": "Record not found"}
    return {"result": result}


@app.get("/data/{record_identifier}/{result_identifier}")
async def retrieve_results(record_identifier: str, result_identifier: str):
    """
    Get specific result given a record identifier and a result identifier
    """
    try:
        result = psm.retrieve(
            record_identifier=record_identifier, result_identifier=result_identifier
        )
    except RecordNotFoundError:
        # TODO this should be more specific than record not found because maybe it's just the result that does not exist
        return {"result": "Record not found"}
    return {"result": result}


@app.get("/output_schema/")
async def retrieve_output_schema(pipeline_type: Optional[str] = None):
    """
    Get the output_schema used by the PipestatManager.
    """

    if pipeline_type == "sample":
        return {"output schema": psm.schema._sample_level_data}
    if pipeline_type == "project":
        return {"output schema": psm.schema._project_level_data}
    if pipeline_type is None:
        return {"output schema": psm.schema}
    else:
        return {"output schema": "output schema not found"}


@app.get("/all_table_contents/")
async def retrieve_table_contents():
    """
    Get all table contents
    """
    # Add skip and limit here as well.
    results = psm.backend.select()
    return {"table_contents": results}


@app.get("/{file_type}/")
async def retrieve_filetype(file_type: str):
    """
    Get all records by filetype
    """
    records_by_filetype = []
    all_records = psm.get_records()["records"]
    for sample in all_records:
        file_result = fetch_pipeline_results(
            project=psm,
            sample_name=sample,
            inclusion_fun=lambda x: x == file_type,
        )
        records_by_filetype.append(file_result)
    return {"records_by_filetype": records_by_filetype}


@app.post("/filtered_table_contents/")
async def retrieve_filtered_table_contents(query_filter: Optional[FilterQuery] = None):
    """
        Get column contents for specific column names and/or filter conditions

    {
      "column_names": [
        "md5sum", "status"
      ],
      "filter_conditions": [["record_identifier", "eq", "random_sample_id2"]]
    }

    """
    try:
        results = psm.backend.select(
            columns=query_filter.column_names,
            filter_conditions=query_filter.filter_conditions,
        )
    except AttributeError:
        return {"response": f"Attribute error for query: {query_filter.column_names}"}
    return {"response": results}


def create_global_pipestatmanager(pipestatcfg):
    """
    build a global pipestatmanager to be used by the endpoints
    """
    global psm
    psm = SamplePipestatManager(config_file=pipestatcfg)
    print("GLOBAL PSM CREATED")


if __name__ != "__main__":
    if os.environ.get("PIPESTAT_CONFIG") is not None:
        pipestatcfg = os.environ.get("PIPESTAT_CONFIG")
        create_global_pipestatmanager(pipestatcfg)
    else:
        _LOGGER.error("Configure by setting PIPESTAT_CONFIG env var")


def main(
    configfile: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
):
    """
    passes relevant info to create a global pipestat manager and then utilizes uvicorn to run at the host address and
    port. These parameters are passed during `pipestat serve` cli usage.
    param: str configfile: a path to a pipestat configfile
    param: str host: host address for uvicorn server
    param: str port: port number for uvicorn server

    """
    pipestatcfg = configfile or os.environ.get("PIPESTAT_CONFIG")
    if pipestatcfg is None:
        _LOGGER.error(
            "A pipestat configuration file must be supplied."
            "Configure by setting PIPESTAT_CONFIG env var or pass as input argument using --config"
        )
    create_global_pipestatmanager(pipestatcfg)
    # Note input argument app vs "reader:app" causes different behavior when using uvicorn.run
    uvicorn.run(
        app,
        host=host or "0.0.0.0",
        port=port or 8001,
    )


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import os
import sys

from setuptools import setup

PACKAGE = "pipestat"

# Additional keyword arguments for setup().
extra = {}

# Ordinary dependencies
DEPENDENCIES = []
with open("requirements/requirements-all.txt", "r") as reqs_file:
    for line in reqs_file:
        if not line.strip():
            continue
        # DEPENDENCIES.append(line.split("=")[0].rstrip("<>"))
        DEPENDENCIES.append(line)

extra["install_requires"] = DEPENDENCIES


# Optional dependencies
# Extras requires a dictionary and not a list?
OPT_DEPENDENCIES = {}
with open("requirements/requirements-db-backend.txt", "r") as reqs_file:
    lines = []
    for line in reqs_file:
        if not line.strip():
            continue
        # OPT_DEPENDENCIES.update({str(line.strip()):line.strip()})
        lines.append(line.strip())
    OPT_DEPENDENCIES.update({"dbbackend": lines})

with open("requirements/requirements-pipestatreader.txt", "r") as reqs_file:
    lines = []
    for line in reqs_file:
        if not line.strip():
            continue
        # OPT_DEPENDENCIES.update({str(line.strip()):line.strip()})
        lines.append(line.strip())
    OPT_DEPENDENCIES.update({"pipestatreader": lines})

extra["extras_require"] = OPT_DEPENDENCIES


# Additional files to include with package
def get_static(name, condition=None):
    static = [
        os.path.join(name, f)
        for f in os.listdir(os.path.join(os.path.dirname(os.path.realpath(__file__)), name))
    ]
    if condition is None:
        return static
    else:
        return [i for i in filter(lambda x: eval(condition), static)]


# scripts to be added to the $PATH
# scripts = get_static("scripts", condition="'.' in x")
# scripts removed (TO remove this)
scripts = None

with open(PACKAGE + "/_version.py", "r") as versionfile:
    version = versionfile.readline().split()[-1].strip("\"'\n")

# Handle the pypi README formatting.
try:
    import pypandoc

    long_description = pypandoc.convert_file("README.md", "rst")
except (IOError, ImportError, OSError):
    long_description = open("README.md").read()

setup(
    name=PACKAGE,
    packages=[PACKAGE],
    version=version,
    description="A pipeline results reporter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="project, metadata, bioinformatics, sequencing, ngs, workflow",
    url="https://github.com/pepkit/" + PACKAGE,
    author="Michal Stolarczyk, Nathan Sheffield",
    license="BSD2",
    entry_points={
        "console_scripts": ["pipestat = pipestat.__main__:main"],
    },
    scripts=scripts,
    include_package_data=True,
    test_suite="tests",
    tests_require=(["mock", "pytest"]),
    setup_requires=(["pytest-runner"] if {"test", "pytest", "ptr"} & set(sys.argv) else []),
    **extra,
)
